#
# Copyright (c) 2012-2014 LAAS/CNRS
# All rights reserved.
#
# Redistribution  and  use  in  source  and binary  forms,  with  or  without
# modification, are permitted provided that the following conditions are met:
#
#   1. Redistributions of  source  code must retain the  above copyright
#      notice and this list of conditions.
#   2. Redistributions in binary form must reproduce the above copyright
#      notice and  this list of  conditions in the  documentation and/or
#      other materials provided with the distribution.
#
# THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
# WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
# MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
# ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
# WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
# IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#
#                                           Anthony Mallet on Mon Mar 12 2012
#

# Pocolibs C client library generation.

template usage {*}{
    "\n"
    "C client library for pocolibs-based components.\n"
    "\n"
    "This template generates a C client library for GenoM component built\n"
    "with the pocolibs middleware.\n"
    "\n"
    "Supported options:\n"
    "  -C, --directory=dir\toutput files in dir\n"
    "  -p, --preserve\tdo not overwrite existing files\n"
    "  -h. --help\t\tprint usage summary (this text)"
}

# defaults
set odir client/c
engine mode +overwrite +move-if-change

# parse options
template options {
    -C - --directory	{ set odir [template arg] }
    -p - --preserve	{ engine mode -overwrite }
    -h - --help		{ engine mode +verbose; puts [template usage]; exit 0 }
}
if {![llength $argv]} { puts [template usage]; exit 2 }

engine chdir $odir

# add common services to dotgen specification
dotgen parse file [file join [dotgen template builtindir] common/genom.gen]

# parse input
set input [list]
foreach f $argv {
  dotgen parse file $f
  lappend input [file normalize $f]
}

# only one component is allowed
set comp [dotgen components]
if {[llength $comp] > 1} {
  template fatal "component compositing not supported"
}
set types [$comp types public]

# require utility procs
template require docutil.tcl

# common header for all files (copyright info from .gen file)
set header {/* <"[--- Generated by [dotgen genom version]. Do not edit - ]"> */

<'if {![catch {dotgen input notice} notice]} {
  puts [lang c; comment $notice]
}'>
}

# client source files
set client_files {
  ../../common/msglib.h
  init.c
  client.c
  msglib.c
  portlib.c
  json.c
  plugin.c
}
foreach c [dotgen components] {
  template parse args [list $c $types] string $header			\
      file [file join [dotgen template builtindir] common/client.h]	\
      file src/[$c name]/c/client.h
  template parse args [list $types] string $header			\
      file [file join [dotgen template builtindir] common/json-types.h]	\
      file src/[$c name]_json-types.h
  foreach f $client_files {
    template parse args $c string $header \
        file $f file src/[$c name]_[file tail $f]
  }
}

template parse args [list $types] \
    string $header file ../../common/serialize.h file src/serialize.h
template parse args [list $types] \
    string $header file ../../common/typecopy.h file src/typecopy.h

# setup build environment
template parse args [list $input] file client.Makefile.am file Makefile.am
template parse file client.configure.ac file configure.ac
template parse file \
    ../../common/autoconf/ax_pthread.m4 file autoconf/ax_pthread.m4
template parse args pocolibs \
    file [file join [dotgen template builtindir] common/c-client.pc.in] \
    file [$comp name]-c-client.pc.in
template parse args pocolibs \
    file [file join \
              [dotgen template builtindir] common/c-client-uninstalled.pc.in] \
    file [$comp name]-c-client-uninstalled.pc.in
template link private.h src/private.h

set deps [list]
foreach d [concat [dotgen input deps] [template deps]] {
  lappend deps "regen: $d"
  lappend deps "$d:"
}
engine mode +overwrite -move-if-change
template parse raw [join $deps "\n"]\n file regen
