<'
# Copyright (c) 2011-2014 LAAS/CNRS
# All rights reserved.
#
# Redistribution and use  in source  and binary  forms,  with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of  source  code must retain the  above copyright
#      notice and this list of conditions.
#   2. Redistributions in binary form must reproduce the above copyright
#      notice and  this list of  conditions in the  documentation and/or
#      other materials provided with the distribution.
#
#                                      Anthony Mallet on Thu Mar  2 2012
#

if {[llength $argv] != 1} { error "expected arguments: component" }
lassign $argv component

# compute handy shortcuts
set comp [$component name]
set COMP [string toupper [$component name]]

lang c
'>
#ifndef H_<"$COMP">_ACTIVITY
#define H_<"$COMP">_ACTIVITY

#include "typecopy.h"
#include "<"$comp">_parameters.h"

/* forward declaration */
struct genom_component_data;

#define <"$COMP">_MAX_ACTIVITIES	32

/* Activities start in the START state. The control task does START->RUNNING
 * when all incompatible activities are VOID, ETHER or STOPPED (intermediate
 * reply). The exec task does RUNNING->ETHER when a codel returns ether. Then
 * the control task then does ETHER->VOID (final reply). If another
 * incompatible activity must run, the control task does RUNNING->STOP. The
 * exec task does STOP->STOPPING and then STOPPING->STOPPED when a codel return
 * ether. The control task then does STOPPED->VOID ("interrupted_by" final
 * reply). */
enum genom_activity_status {
   ACT_VOID,
   ACT_START, ACT_STOP,
   ACT_RUNNING, ACT_STOPPING,
   ACT_ETHER, ACT_STOPPED
};

struct genom_activity {
  long taskid;
  int sid, rid, aid;
  enum genom_activity_status status;
  genom_event codel;
  void *exdetail;
  const char *interruptedby;
  int stateid;
};


/* === Service activity ==================================================== */

<'foreach s [$component services] {'>

/* <"[--- [$s name] -----------------------------------------------------]"> */

/* activity data */
struct genom_<"$comp">_<"[$s name]">_activity {
  struct genom_activity h;
  struct genom_<"$comp">_<"[$s name]">_input in;
  struct genom_<"$comp">_<"[$s name]">_output out;
<'  if {[llength [$s parameters local]]} {'>
  struct genom_<"$comp">_<"[$s name]">_locals locals;
<'  }'>
};

/* input/output initializer */
static __inline__ void
genom_tinit_<"$comp">_<"[$s name]">_activity(
  struct genom_<"$comp">_<"[$s name]">_activity *a)
{
<'  foreach p [$s parameters in inout] {'>
  genom_tinit_<"[[$p type] mangle]">(
    <"[[$p type] pass reference a->in.[$p name]]">);
<'  }'>
<'  foreach p [$s parameters inout out] {'>
  genom_tinit_<"[[$p type] mangle]">(
    <"[[$p type] pass reference a->out.[$p name]]">);
<'  }'>
<'  foreach p [$s parameters local] {'>
  genom_tinit_<"[[$p type] mangle]">(
    <"[[$p type] pass reference a->locals.[$p name]]">);
<'  }'>
}

/* input/output finalizer */
static __inline__ void
genom_tfini_<"$comp">_<"[$s name]">_activity(
  struct genom_<"$comp">_<"[$s name]">_activity *a)
{
<'  foreach p [$s parameters in inout] {'>
  genom_tfini_<"[[$p type] mangle]">(
    <"[[$p type] pass reference a->in.[$p name]]">);
<'  }'>
<'  foreach p [$s parameters inout out] {'>
  genom_tfini_<"[[$p type] mangle]">(
    <"[[$p type] pass reference a->out.[$p name]]">);
<'  }'>
<'  foreach p [$s parameters local] {'>
  genom_tfini_<"[[$p type] mangle]">(
    <"[[$p type] pass reference a->locals.[$p name]]">);
<'  }'>
}

<'}'>


/* === Activities ========================================================== */

struct genom_activities {
  int last;
  union {
    struct genom_activity any;
<'foreach s [$component services] {'>
    struct genom_<"$comp">_<"[$s name]">_activity s_<"[$s name]">;
<'}'>
  } i[<"$COMP">_MAX_ACTIVITIES];
};

int	genom_<"$comp">_activity_alloc(struct genom_component_data *self);
void	genom_<"$comp">_activity_free(struct genom_component_data *self, int id);
void	genom_<"$comp">_activity_report(struct genom_component_data *self,
		struct genom_activity *a);
int	genom_<"$comp">_activity_encodex(char *buffer, int size, char *dst,
                int maxsize);
int	genom_<"$comp">_interrupt_reqd(struct genom_component_data *self,
		struct genom_activity *a);

#endif /* H_<"$COMP">_ACTIVITY */
