<'
# Copyright (c) 2011-2014 LAAS/CNRS
# All rights reserved.
#
# Redistribution and use  in source  and binary  forms,  with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of  source  code must retain the  above copyright
#      notice and this list of conditions.
#   2. Redistributions in binary form must reproduce the above copyright
#      notice and  this list of  conditions in the  documentation and/or
#      other materials provided with the distribution.
#
#                                      Anthony Mallet on Wed Feb 29 2012
#

if {[llength $argv] != 1} { error "expected arguments: component" }
lassign $argv component

# compute handy shortcuts
set comp [$component name]

lang c
'>
#include "autoconf/acheader.h"

#include <sys/types.h>
#include <signal.h>
#include <stdio.h>
#include <unistd.h>

#include "commonStructLib.h"
#include "h2evnLib.h"
#include "taskLib.h"

#include "<"$comp">_parameters.h"
#include "<"$comp">_msglib.h"
#include "<"$comp">_activity.h"
#include "<"$comp">_control_task.h"
#include "typecopy.h"


/* --- local data ---------------------------------------------------------- */

#define CNTRL_TASK_MIN_STACK_SIZE 4096
#define EXEC_TASK_MIN_STACK_SIZE  4096

/* cslib server with context pointer, for reentrant csLib callbacks */
struct CS_SERV_IDS {
  CS_SERV csserv;
  struct genom_component_data *self;
};

<'foreach s [$component services] {'>
static void	<"$comp">_<"[$s name]">_rqstcb(SERV_ID csserv, int sid);
<'}'>
static void *	<"$comp">_cntrl_task(void *);


/* <"[--- genom_${comp}_init --------------------------------------------]"> */

void *
genom_<"$comp">_init(void)
{
  struct genom_component_data *self;
  char tname[64];
  STATUS s;
  int i;

  /* create internal data structure */
  if (commonStructCreate(sizeof(*self), (void **)&self) != OK)  {
    genom_log_warn(1, "cannot create internal data structure");
    return NULL;
  }
  self->context.raise = genom_pocolibs_raise;
  self->context.raised = genom_pocolibs_raised;

  self->context.data = &self->context_data;
  self->context_data.self = self;
  self->context_data.ex = genom_ok;
  self->context_data.exdetail = NULL;
  self->context_data.exsize = 0;

  self->shutdown = 0;
  snprintf(self->mbox_name, sizeof(self->mbox_name), "%s", genom_instance);
  self->csserv = NULL;
  self->taskid = ERROR;
  self->sync = semBCreate(SEM_Q_PRIORITY, SEM_EMPTY);
  if (!self->sync) {
    genom_log_warn(1, "cannot create control semaphore");
    goto error;
  }

<'if {![catch {$component ids} ids]} {'>
  genom_tinit_<"[$ids mangle]">(&self->ids);
<'}'>
<'foreach p [$component ports] {'>
  genom_tinit_<"$comp">_<"[$p name]">_port(&self->ports.<"[$p name]">);
<'}'>
<'foreach r [$component remotes] {'>
  genom_tinit_<"$comp">_<"[$r name]">_remote(&self->remotes.<"[$r name]">);
<'}'>
<'foreach t [$component tasks] { '>
  self->tasks.<"[$t name]">.context.raise = genom_pocolibs_raise;
  self->tasks.<"[$t name]">.context.raised = genom_pocolibs_raised;
  self->tasks.<"[$t name]">.context.data =
    &self->tasks.<"[$t name]">.context_data;

  self->tasks.<"[$t name]">.context_data.self = self;
  self->tasks.<"[$t name]">.context_data.ex = genom_ok;
  self->tasks.<"[$t name]">.context_data.exdetail = NULL;
  self->tasks.<"[$t name]">.context_data.exsize = 0;

  self->tasks.<"[$t name]">.taskid = ERROR;
  self->tasks.<"[$t name]">.runnable = 0;
  self->tasks.<"[$t name]">.sync = semBCreate(SEM_Q_PRIORITY, SEM_EMPTY);
  if (!self->tasks.<"[$t name]">.sync) {
    genom_log_warn(1, "cannot create control semaphore");
    goto error;
  }
<'}'>

  self->activities.last = 0;
  for(i=0; i<sizeof(self->activities.i)/sizeof(self->activities.i[0]); i++) {
    self->activities.i[i].any.status = ACT_VOID;
  }
  memset(self->run_map, 0, sizeof(self->run_map));

  /* create posters */
  s = genom_metadata_<"$comp">_init(self);
  if (s) goto error;
<'foreach p [$component ports simple] {'>
  if (genom_<"$comp">_<"[$p name]">_open(&self->context)) goto error;
<'}'>
  s = genom_state_<"$comp">_init(self);
  if (s) goto error;
  if (genom_<"$comp">_genom_metadata_write(&self->context)) goto error;


  /* spawn execution tasks */
<'foreach t [$component tasks] {'>
<'  if {[catch {[$t priority] value} priority]} { set priority 255 }'>
<'  if {[catch {[$t stack] value} stack]} { set stack [expr {1<<23}] }'>
  snprintf(tname, sizeof(tname), "%s/%s", genom_instance, "<"[$t name]">");
  if (taskSpawn2(tname, <"$priority">,
                 VX_FP_TASK, EXEC_TASK_MIN_STACK_SIZE + <"$stack">,
                 genom_<"$comp">_<"[$t name]">_exec_task, self) == ERROR) {
    genom_log_warn(1, "cannot spawn %s exec task", "<"[$t name]">");
    goto error;
  }
  s = semTake(self->tasks.<"[$t name]">.sync, WAIT_FOREVER);
  s = commonStructTake(self);
  if (self->tasks.<"[$t name]">.taskid == ERROR) {
    genom_log_warn(0, "%s exec task failed", "<"[$t name]">");
    s = commonStructGive(self);
    goto error;
  }
  s = commonStructGive(self);
<'}'>

  /* spawn control task */
  snprintf(tname, sizeof(tname), "%s", genom_instance);
  if (taskSpawn2(tname, 10 /* priority */,
                 VX_FP_TASK, <"$COMP">_MAX_RQST_SIZE + <"$COMP">_MAX_REPLY_SIZE
                 + CNTRL_TASK_MIN_STACK_SIZE /*size*/,
                 <"$comp">_cntrl_task, self) == ERROR) {
    genom_log_warn(1, "cannot spawn control task");
    goto error;
  }
  s = semTake(self->sync, WAIT_FOREVER);
  s = commonStructTake(self);
  if (self->taskid == ERROR) {
    genom_log_warn(0, "control task failed");
    s = commonStructGive(self);
    goto error;
  }
  s = commonStructGive(self);

  genom_log_info("setup and running");
  return self;

error:
  genom_<"$comp">_fini(self);
  return NULL;
}


/* <"[--- genom_${comp}_fini --------------------------------------------]"> */

void
genom_<"$comp">_fini(void *data)
{
  struct genom_component_data *self = data;
  STATUS s;

  s = commonStructTake(self);
  if (s != OK) assert(!"cannot lock ids - giving up");
  self->shutdown = 1;
  genom_<"$comp">_schedule_exec(self);
  genom_<"$comp">_schedule_cntrl(self);
  s = commonStructGive(self);

<'foreach t [$component tasks] {'>
  if (self->tasks.<"[$t name]">.taskid != ERROR) {
    s = semTake(self->tasks.<"[$t name]">.sync, WAIT_FOREVER);
  }
<'}'>
  if (self->taskid != ERROR) {
    s = semTake(self->sync, WAIT_FOREVER);
  }

<'foreach t [$component tasks] {'>
  if (self->tasks.<"[$t name]">.sync)
    s = semDelete(self->tasks.<"[$t name]">.sync);
<'}'>
  if (self->sync)
    s = semDelete(self->sync);

<'foreach p [$component ports] {'>
  genom_<"$comp">_<"[$p name]">_delete(&self->context);
<'}'>

<'foreach r [$component remotes] {'>
  genom_tfini_<"$comp">_<"[$r name]">_remote(&self->remotes.<"[$r name]">);
<'}'>
<'foreach p [$component ports] {'>
  genom_tfini_<"$comp">_<"[$p name]">_port(&self->ports.<"[$p name]">);
<'}'>
<'if {![catch {$component ids} ids]} {'>
  genom_tfini_<"[$ids mangle]">(&self->ids);
<'}'>
  commonStructDelete(self);
  genom_log_info("shutdown complete");
}


/* <"[--- genom_${comp}_schedule_cntrl ----------------------------------]"> */

void
genom_<"$comp">_schedule_cntrl(struct genom_component_data *self)
{
  if (self->taskid == ERROR) return;
  h2evnSignal(self->taskid);
}


/* <"[--- genom_${comp}_schedule_exec -----------------------------------]"> */

void
genom_<"$comp">_schedule_exec(struct genom_component_data *self)
{
<'foreach t [$component tasks] {'>
  if (self->tasks.<"[$t name]">.taskid != ERROR) {
    if (!self->tasks.<"[$t name]">.runnable) {
<'  if {[catch {$t period}]} {'>
      h2evnSignal(self->tasks.<"[$t name]">.taskid);
<'  }'>
      self->tasks.<"[$t name]">.runnable = 1;
    }
  }
<'}'>
}


/* <"[--- ${comp}_cntrl_task(void) --------------------------------------]"> */

static void *
<"$comp">_cntrl_task(void *data)
{
  struct genom_component_data *self = data;
  genom_state_component *state = &self->ports.genom_state.h.buffer;
  struct genom_activity *a;
  genom_event s;
  int id;
  int e;

  commonStructTake(self);

  /* create requests reception mailbox */
  if (csMboxInit(self->mbox_name, <"$COMP">_MBOX_RQST_SIZE, 0) != OK) {
    genom_log_warn(1, "cannot create server mailbox");
    commonStructGive(self);
    semGive(self->sync);
    return NULL;
  }

  /* configure as a server */
  if (csServInitN(
	<"$COMP">_MAX_RQST_SIZE,
	<"$COMP">_MAX_REPLY_SIZE >= <"$COMP">_MAX_INTERMED_REPLY_SIZE ?
	<"$COMP">_MAX_REPLY_SIZE : <"$COMP">_MAX_INTERMED_REPLY_SIZE,
	<"$COMP">_NRQSTID, &self->csserv) != OK) {
    genom_log_warn(1, "cannot initialize server mailbox");
    goto error;
  }
  /* fiddle with returned CS_SERV structure, to store ids for callbacks */
  self->csserv = realloc(self->csserv, sizeof(struct CS_SERV_IDS));
  ((struct CS_SERV_IDS *)self->csserv)->self = self;

  /* install services callbacks */
<'foreach s [$component services] {'>
  if (csServFuncInstall(self->csserv, <"$COMP">_<"[$s name]">_RQSTID,
                        (FUNCPTR)<"$comp">_<"[$s name]">_rqstcb) != OK) {
    genom_log_warn(1, "cannot serve service <"[$s name]">");
    goto error;
  }
<'}'>

  /* signal parent */
  self->taskid = taskIdSelf();
  genom_log_info("spawned control task");
  semGive(self->sync);

  while(1) {
    commonStructGive(self);
    if (h2evnSusp(0) != TRUE) {
      genom_log_warn(1, "someone did something nasty");
      genom_log_warn(0, "aborting");
      abort();
    }
    commonStructTake(self);
    genom_log_debug("control task wake up");

    /* read reception mailbox status */
    e = csMboxStatus(RCV_MBOX);
    if (e & RCV_MBOX)  {
      if (csServRqstExec(self->csserv) != OK) {
        genom_log_warn(1, "someone did something nasty");
        genom_log_warn(0, "aborting");
        abort();
      }
    }

    /* process pending activities */
    for(e=id=0; id<<"$COMP">_MAX_ACTIVITIES; id++) {
      a = &self->activities.i[id].any;
      switch(a->status) {
        case ACT_VOID:
        case ACT_RUNNING:
        case ACT_STOPPING:
        case ACT_STOP:
          break;
        case ACT_START:
          if (!genom_<"$comp">_interrupt_reqd(self, a)) {
            /* update state port */
            if (a->stateid >= 0) {
              strncpy(state->activity._buffer[a->stateid].state,
                      <"$comp">_start,
                      sizeof(state->activity._buffer[a->stateid].state));
              e = 1;
            }

            a->status = ACT_RUNNING;
            a->codel = <"$comp">_start;
            genom_<"$comp">_schedule_exec(self);
          }
          break;
        case ACT_STOPPED:
        case ACT_ETHER:
          genom_<"$comp">_activity_report(self, a);
          /* wake up sleeping activities */
          genom_<"$comp">_schedule_exec(self);
          break;
      }
    }

    /* update state port */
    if (e) {
      s = genom_<"$comp">_genom_state_write(&self->context);
      if (s) genom_log_warn(0, "cannot update state port: %s", s);
    }

    
    /* handle shutdown: wait for all exec tasks, then break */
    if (self->shutdown) {
<'foreach t [$component tasks] {'>
      if (self->tasks.<"[$t name]">.taskid != ERROR) continue;
<'}'>
      break;
    }
  }

  genom_log_info("shutting down control task");

error:
  if (self->csserv) csServEnd(self->csserv);
  csMboxEnd();
  self->taskid = ERROR;
  commonStructGive(self);
  semGive(self->sync);
  return NULL;
}


/* --- Service callbacks --------------------------------------------------- */

<'foreach s [$component services] {'>

static void
<"$comp">_<"[$s name]">_rqstcb(SERV_ID csserv, int rid)
{
  struct genom_component_data *self = ((struct CS_SERV_IDS *)csserv)->self;
  struct genom_<"$comp">_<"[$s name]">_activity *a;
  STATUS s;
  int delay;
  int id;

  genom_log_debug("handling request for %s", "<"[$s name]">");

  /* get an activity id */
  id = genom_<"$comp">_activity_alloc(self);
  if (id < 0) {
    s = csServReplySend(
      csserv, rid, FINAL_REPLY, ERROR, NULL, 0,
      genom_<"$comp">_genom_too_many_activities_encodex);
    if (s == ERROR) assert(!"unexpected failure");
    return;
  }

  a = &self->activities.i[id].s_<"[$s name]">;
  assert(a->h.status == ACT_VOID);
<'  if {[catch {$s task} t]} {'>
  a->h.taskid = 0;
<'  } else {'>
  a->h.taskid = self->tasks.<"[$t name]">.taskid;
<'  }'>
  a->h.sid = <"$COMP">_<"[$s name]">_RQSTID;
  a->h.rid = rid;
  a->h.aid = id;
  a->h.status = ACT_START;
  genom_tinit_<"$comp">_<"[$s name]">_activity(a);

  /* decode input */
  s = csServRqstParamsGet(csserv, rid, (char *)&a->in, 0,
                          genom_<"$comp">_<"[$s name]">_decode);
  if (s == ERROR) {
    genom_log_warn(0, "invalid input for service %s", "<"[$s name]">");
    s = csServReplySend(
      csserv, rid, FINAL_REPLY, ERROR, NULL, 0,
      genom_<"$comp">_genom_serialization_encodex);
    if (s == ERROR) assert(!"unexpected failure");
    goto error;
  }

  /* control codels (validate + stateless) */
  a->h.codel = genom_<"$comp">_<"[$s name]">_controlcb(self, a);
  if (a->h.codel) {
    a->h.exdetail = (void *)self->context.raised(NULL, &self->context);
    s = csServReplySend(csserv, rid, FINAL_REPLY, ERROR, (char *)&a->h, 0,
                        genom_<"$comp">_activity_encodex);
    if (s == ERROR) assert(!"unexpected failure");
    goto error;
  }

  /* interrupt incompatible activities */
  delay = genom_<"$comp">_interrupt_reqd(self, (struct genom_activity *)a);

<'  if {[catch {$s task} t]} {'>
  /* update after/before array */
  self->run_map[<"$COMP">_<"[$s name]">_RQSTID] = 1;

  /* send final reply */
  s = csServReplySend(csserv, rid, FINAL_REPLY, OK,
                      (char *)&a->out, 0, genom_<"$comp">_<"[$s name]">_encode);
  if (s == ERROR) {
    genom_log_warn(1, "could not send output for service %s", "<"[$s name]">");
    s = csServReplySend(
      csserv, rid, FINAL_REPLY, ERROR, NULL, 0,
      genom_<"$comp">_genom_serialization_encodex);
    if (s == ERROR) {
      genom_log_warn(1, "discarding output for service %s", "<"[$s name]">");
    } else {
      genom_log_warn(0, "invalid output for service %s", "<"[$s name]">");
    }
    goto error;
  }

  if (delay) /* silence brain-dead gcc warning -Wunused-but-set */;

  genom_tfini_<"$comp">_<"[$s name]">_activity(a);
  genom_<"$comp">_activity_free(self, a->h.aid);

  genom_<"$comp">_schedule_exec(self); /* wake up sleeping activities */
<'  } else {'>
  /* send intermediate reply */
  s = csServReplySend(csserv, rid, INTERMED_REPLY, OK,
                      (char *)&a->h.aid, sizeof(int), NULL);
  if (s == ERROR) {
    genom_log_warn(1, "cannot acknowledge service %s", "<"[$s name]">");
  }

  /* update state port */
  {
    genom_event e;
    genom_state_component *state = &self->ports.genom_state.h.buffer;
    if (state->activity._length < state->activity._maximum) {
      a->h.stateid = state->activity._length++;
      state->activity._buffer[a->h.stateid].id = a->h.aid;
      strncpy(state->activity._buffer[a->h.stateid].name, "<"[$s name]">",
              sizeof(state->activity._buffer[a->h.stateid].name));
      if (delay)
        state->activity._buffer[a->h.stateid].state[0] = *"";
      else
        strncpy(state->activity._buffer[a->h.stateid].state, <"$comp">_start,
                sizeof(state->activity._buffer[a->h.stateid].state));

      e = genom_<"$comp">_genom_state_write(&self->context);
      if (e) genom_log_warn(0, "cannot update state port: %s", e);
    } else
      a->h.stateid = -1;
  }

  if (!delay) {
    /* schedule now */
    a->h.status = ACT_RUNNING;
    a->h.codel = <"$comp">_start;
    genom_<"$comp">_schedule_exec(self);
  }
<'  }'>

  return;

error:
  genom_tfini_<"$comp">_<"[$s name]">_activity(a);
  genom_<"$comp">_activity_free(self, a->h.aid);
}

<'}'>
