<'
# Copyright (c) 2012-2014 LAAS/CNRS
# All rights reserved.
#
# Redistribution and use  in source  and binary  forms,  with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of  source  code must retain the  above copyright
#      notice and this list of conditions.
#   2. Redistributions in binary form must reproduce the above copyright
#      notice and  this list of  conditions in the  documentation and/or
#      other materials provided with the distribution.
#
#                                      Anthony Mallet on Wed Aug  1 2012
#

if {[llength $argv] != 2} { error "expected arguments: component task" }
lassign $argv component task

# compute handy shortcuts
set comp [$component name]
set tname [$task name]

lang c
'>
#include "autoconf/acheader.h"

#include "commonStructLib.h"

#include "<"$comp">_control_task.h"


/* === task codels ========================================================= */
<'
foreach e [$task fsm] {
  set codel [$task fsm $e]
'>

/* state <"[$e cname]">  */
static __inline__ genom_event
<"$comp">_codel_<"$tname">_<"[$e cname]">(struct genom_component_data *self)
{
  genom_event s;

<'  # copy parameters for async codels
  if {[$codel kind] == "async"} {
    foreach p [$codel parameters ids] {'>
  <"[[$p type] declarator [$p name]_]">;
<'    }'>

<'    foreach p [$codel parameters ids] {'>
  genom_tinit_<"[[$p type] mangle]">(<"[[$p type] pass reference [$p name]_]">);
  genom_tcopy_<"[[$p type] mangle]">(
    <"[[$p type] pass reference [$p name]_]">,
    <"[[$p type] pass value self->ids[$p member]]">);
<'    }'>
<'  } '>

<'    # build parameter list
  set plist [list]
  foreach p [$codel parameters] {
    switch -glob -- "[$codel kind]/[$p src]" {
      */local	{ error "task codel with service parameter" }
      */remote	{ set m "self->remotes.[[$p remote] name].data" }
      */port	{ set m "self->ports.[[$p port] name].handle" }
      async/ids	{ set m [$p name]_ }
      sync/ids	{ set m "self->ids[$p member]" }
      default	{ error "reached unreachable code" }
    }

    switch  -glob -- "[$p dir]/[$p src]" {
      in/* - */port {
        lappend plist [[$p type] pass value $m]
      }
      default {
        lappend plist [[$p type] pass reference $m]
      }
    }
  }
  lappend plist "&self->tasks.${tname}.context"
'>

<'if {[$codel kind] == "async"} {'>
  commonStructGive(self);
<'}'>
  s = <"[$codel invoke $plist]">;
<'if {[$codel kind] == "async"} {'>
  commonStructTake(self);

<'  foreach p [$codel parameters ids] {'>
<'    if {[$p dir] != "in"} {'>
  genom_tcopy_<"[[$p type] mangle]">(
    <"[[$p type] pass reference self->ids[$p member]]">,
    <"[[$p type] pass value [$p name]_]">);
<'    }'>
  genom_tfini_<"[[$p type] mangle]">(<"[[$p type] pass reference [$p name]_]">);
<'  }'>
<'}'>
  return s;
}
<'}'>


/* <"[--- genom_${comp}_${tname}_invoke_perm ----------------------------]"> */

genom_event
genom_<"$comp">_<"$tname">_invoke_perm(struct genom_component_data *self)
{
<'if {[llength [$task fsm]]} {'>
  genom_event s;
<'}'>
  if (self->tasks.<"$tname">.pact == <"$comp">_ether)
    return <"$comp">_ether;

  genom_log_debug("task %s invoking permanent activity", "<"$tname">");

<'foreach e [$task fsm] {'>
  if (self->tasks.<"$tname">.pact == <"[$e cname]">) {
    s = <"$comp">_codel_<"$tname">_<"[$e cname]">(self);
    genom_log_debug("permanent activity in task %s yielded %s",
                    "<"$tname">", s);
    if (
<'  foreach y [[$task fsm $e] yield] {'>
      s == <"[$y cname]"> ||
<'  }'>
      0) {
      return s;
    }
    goto ex;
  }
<'}'>

  /* default events, if not already handled above */
  if (self->tasks.<"$tname">.pact == <"$comp">_sleep) return <"$comp">_sleep;
  if (self->tasks.<"$tname">.pact == <"$comp">_start ||
      self->tasks.<"$tname">.pact == <"$comp">_stop)
    return <"$comp">_ether;

  /* this cannot happen by construction */
  genom_log_warn(0, "bad state %s in task %s permanent activity",
                 self->tasks.<"$tname">.pact, "<"$tname">");
  assert(!"bad state");
  abort();
  return NULL;

<'if {[llength [$task fsm]]} {'>
ex:
  if (
<'  foreach t [$task throw] {'>
    s != <"[$t cname]">_id &&
<'  }'>
    1) {
    genom_bad_transition_detail d;
    strncpy(d.from, self->tasks.<"$tname">.pact, sizeof(d.from));
    d.from[sizeof(d.from)-1] = *"";
    strncpy(d.to, s, sizeof(d.to));
    d.to[sizeof(d.to)-1] = *"";
    genom_log_warn(
      0, "bad transition from %s to %s in task %s",
        self->tasks.<"$tname">.pact, s, "<"$tname">");
    self->tasks.<"$tname">.status =
      genom_bad_transition(&d, &self->tasks.<"$tname">.context);
    return <"$comp">_ether;
  }

  self->tasks.<"$tname">.status = s;
  return <"$comp">_ether;
<'}'>
}


/* === services codels ===================================================== */

<'foreach s [$task services] {'>
/* <"[--- Service [$s name] ---------------------------------------------]"> */

<'  foreach e [$s fsm] { set codel [$s fsm $e]'>

/* state <"[$e cname]"> */
static __inline__ genom_event
<"$comp">_<"[$s name]">_codel_<"[$e cname]">(
  struct genom_component_data *self,
  struct genom_<"$comp">_<"[$s name]">_activity *a)
{
  genom_event s;

<'  # copy parameters for async codels
  if {[$codel kind] == "async"} {
    foreach p [$codel parameters ids] {'>
  <"[[$p type] declarator [$p name]_]">;
<'    }'>

<'    foreach p [$codel parameters ids] {'>
  genom_tinit_<"[[$p type] mangle]">(<"[[$p type] pass reference [$p name]_]">);
  genom_tcopy_<"[[$p type] mangle]">(
    <"[[$p type] pass reference [$p name]_]">,
    <"[[$p type] pass value self->ids[$p member]]">);
<'    }'>
<'  } '>

<'    # build parameter list
    set plist [list]
    foreach p [$codel parameters] {
      switch -glob -- "[$codel kind]/[$p src]/[$p dir]" {
        */local/*	{
          switch -- [[$p param] dir] {
            local	{ set m "a->locals.[$p name]" }
            in		{ set m "a->in.[$p name]" }
            inout - out { set m "a->out.[$p name]" }
            default	{ error "unsupported local direction" }
          }
        }
        */remote/*	{ set m "self->remotes.[[$p remote] name].data" }
        */port/*	{ set m "self->ports.[[$p port] name].handle" }
        async/ids/*	{ set m [$p name]_ }
        sync/ids/*	{ set m "self->ids[$p member]" }
        default		{ error "reached unreachable code" }
      }

      switch  -glob -- "[$p dir]/[$p src]" {
	in/* - */port {
          lappend plist [[$p type] pass value $m]
        }
	default {
	  lappend plist [[$p type] pass reference $m]
	}
      }
    }
    lappend plist "&self->tasks.${tname}.context"
'>
<'    if {[$codel kind] == "async"} {'>
  commonStructGive(self);
<'    }'>
  s = <"[$codel invoke $plist]">;
<'    if {[$codel kind] == "async"} {'>
  commonStructTake(self);

<'      foreach p [$codel parameters ids] {'>
<'        if {[$p dir] != "in"} {'>
  genom_tcopy_<"[[$p type] mangle]">(
    <"[[$p type] pass reference self->ids[$p member]]">,
    <"[[$p type] pass value [$p name]_]">);
<'        }'>
  genom_tfini_<"[[$p type] mangle]">(<"[[$p type] pass reference [$p name]_]">);
<'      }'>
<'    }'>
  return s;
}
<'  }'>

/* invoke codels according to <"[$s name]"> state */
static __inline__ genom_event
<"$comp">_<"$tname">_<"[$s name]">_invoke(
  struct genom_component_data *self, struct genom_activity *a)
{
  genom_event s;

  genom_log_debug(
    "task %s invoking activity %s", "<"$tname">", "<"[$s name]">");

<'  foreach e [$s fsm] {'>
  if (a->codel == <"[$e cname]">) {
    s = <"$comp">_<"[$s name]">_codel_<"[$e cname]">(
      self, (struct genom_<"$comp">_<"[$s name]">_activity *)a);
    genom_log_debug("service %s in task %s yielded %s",
                    "<"[$s name]">", "<"$tname">", s);
    if(
<'    foreach y [[$s fsm $e] yield] {'>
      s == <"[$y cname]"> ||
<'    }'>
      0) {
      return s;
    }
    goto ex;
  }

<'  }'>
  /* default events, if not already handled above */
  if (a->codel == <"$comp">_sleep) return <"$comp">_sleep;
  if (a->codel == <"$comp">_stop) return <"$comp">_ether;

  /* this cannot happen by construction */
  genom_log_warn(0, "bad state %s in service %s", a->codel, "<"[$s name]">");
  assert(!"bad state");
  abort();
  return NULL;

ex:
  if (
<'  foreach t [$s throw] {'>
    s != <"[$t cname]">_id &&
<'  }'>
    1) {
    genom_bad_transition_detail d;
    strncpy(d.from, a->codel, sizeof(d.from)); d.from[sizeof(d.from)-1] = *"";
    strncpy(d.to, s, sizeof(d.to)); d.to[sizeof(d.to)-1] = *"";
    genom_log_warn(0, "bad transition from %s to %s in service %s",
                   a->codel, s?s:"genom_ok", "<"[$s name]">");
    genom_bad_transition(&d, &self->tasks.<"$tname">.context);
  }

  return NULL;
}

<'}'>

/* <"[--- genom_${comp}_${tname}_invoke ---------------------------------]"> */

genom_event
genom_<"$comp">_<"$tname">_invoke(
  struct genom_component_data *self, struct genom_activity *a)
{
  switch(a->sid) {
<'  foreach s [$task services] {'>
    case <"$COMP">_<"[$s name]">_RQSTID:
      return <"$comp">_<"$tname">_<"[$s name]">_invoke(self, a);
<'  }'>
  }

  /* this cannot happen by construction */
  genom_log_warn(0, "bad activity %d in task %s", a->sid, "<"$tname">");
  assert(!"bad activity");
  abort();
  return NULL;
}
